@@ -16,10 +16,14 @@ use pretty_hex::PrettyHex;
 
 pub(crate) mod aac;
 pub(crate) mod g723;
-pub(crate) mod h264;
+
+#[doc(hidden)]
+pub mod h264;
+
 pub(crate) mod onvif;
 pub(crate) mod simple_audio;
 
+#[derive(Debug)]
 pub enum CodecItem {
     VideoFrame(VideoFrame),
     AudioFrame(AudioFrame),
@@ -259,25 +263,29 @@ pub struct VideoFrame {
     /// In H.264 terms, this is a frame with `nal_ref_idc == 0`.
     pub is_disposable: bool,
 
-    /// Position within `concat(data_prefix, data)`.
-    pos: u32,
-
-    data_prefix: [u8; 4],
-
-    /// Frame content in the requested format. Currently in a single [bytes::Bytes]
-    /// allocation, but this may change when supporting H.264 partitioned slices
-    /// or if we revise the fragmentation implementation.
     data: bytes::Bytes,
 }
 
 impl VideoFrame {
+    #[inline]
     pub fn start_ctx(&self) -> crate::Context {
         self.start_ctx
     }
 
+    #[inline]
     pub fn end_ctx(&self) -> crate::Context {
         self.end_ctx
     }
+
+    #[inline]
+    pub fn data(&self) -> &Bytes {
+        &self.data
+    }
+
+    #[inline]
+    pub fn into_data(self) -> Bytes {
+        self.data
+    }
 }
 
 impl std::fmt::Debug for VideoFrame {
@@ -291,65 +299,24 @@ impl std::fmt::Debug for VideoFrame {
             .field("new_parameters", &self.new_parameters)
             .field("is_random_access_point", &self.is_random_access_point)
             .field("is_disposable", &self.is_disposable)
-            .field("pos", &self.pos)
-            .field("data_len", &(self.data.len() + 4))
+            .field("data_len", &self.data.len())
             //.field("data", &self.data.hex_dump())
             .finish()
     }
 }
 
-impl bytes::Buf for VideoFrame {
-    fn remaining(&self) -> usize {
-        self.data.len() + 4 - (self.pos as usize)
-    }
-
-    fn chunk(&self) -> &[u8] {
-        let pos = self.pos as usize;
-        if let Some(pos_within_data) = pos.checked_sub(4) {
-            &self.data[pos_within_data..]
-        } else {
-            &self.data_prefix[pos..]
-        }
-    }
-
-    fn advance(&mut self, cnt: usize) {
-        assert!((self.pos as usize) + cnt <= 4 + self.data.len());
-        self.pos += cnt as u32;
-    }
-
-    fn chunks_vectored<'a>(&'a self, dst: &mut [std::io::IoSlice<'a>]) -> usize {
-        match dst.len() {
-            0 => 0,
-            1 => {
-                dst[0] = std::io::IoSlice::new(self.chunk());
-                1
-            }
-            _ if self.pos < 4 => {
-                dst[0] = std::io::IoSlice::new(&self.data_prefix[self.pos as usize..]);
-                dst[1] = std::io::IoSlice::new(&self.data);
-                2
-            }
-            _ => {
-                dst[0] = std::io::IoSlice::new(&self.data[(self.pos - 4) as usize..]);
-                1
-            }
-        }
-    }
-}
-
 /// Turns RTP packets into [CodecItem]s.
-/// This interface unstable and for internal use; it's exposed for direct fuzzing.
+/// This interface unstable and for internal use; it's exposed for direct fuzzing and benchmarking.
 #[doc(hidden)]
 #[derive(Debug)]
 pub struct Depacketizer(DepacketizerInner);
 
 #[derive(Debug)]
-#[allow(clippy::clippy::large_enum_variant)]
 enum DepacketizerInner {
-    Aac(aac::Depacketizer),
+    Aac(Box<aac::Depacketizer>),
     SimpleAudio(simple_audio::Depacketizer),
     G723(g723::Depacketizer),
-    H264(h264::Depacketizer),
+    H264(Box<h264::Depacketizer>),
     Onvif(onvif::Depacketizer),
 }
 
@@ -366,15 +333,15 @@ impl Depacketizer {
         // RTP Payload Format Media Types
         // https://www.iana.org/assignments/rtp-parameters/rtp-parameters.xhtml#rtp-parameters-2
         Ok(Depacketizer(match (media, encoding_name) {
-            ("video", "h264") => DepacketizerInner::H264(h264::Depacketizer::new(
+            ("video", "h264") => DepacketizerInner::H264(Box::new(h264::Depacketizer::new(
                 clock_rate,
                 format_specific_params,
-            )?),
-            ("audio", "mpeg4-generic") => DepacketizerInner::Aac(aac::Depacketizer::new(
+            )?)),
+            ("audio", "mpeg4-generic") => DepacketizerInner::Aac(Box::new(aac::Depacketizer::new(
                 clock_rate,
                 channels,
                 format_specific_params,
-            )?),
+            )?)),
             ("audio", "g726-16") => {
                 DepacketizerInner::SimpleAudio(simple_audio::Depacketizer::new(clock_rate, 2))
             }
